/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: display.c
 * Author: Jed Marti
 * Description: Display protoBuf created structures, These look sort of like JSON but are more
 *   Lisp like.
 * Created: Fri Dec  6 11:34:48 2024
 * Revision History:
 *   Sun Dec 29 10:46:58 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>

#include "protobuf.h"


// Display functions. Functions set up during initialization. The display is table driven by
// dynamic types. If you add a new type, increase MAXTYPES (see protobuf.h). The new type should
// have an entry in this table to show it.
//  Your display function must take two arguments:
//     fh: The file handle to display to.
//     g: The generic value - guaranteed by the caller to be your type.
displayFn displayFns[MAXTYPES];


static void illegalType(FILE *fh, Generic *g)
/*     Shouldn't happen but keeps us from undefined problems. Causes halt.
 * Parameters:
 *   fh: File to write to.
 *   g: Address of corrupted to display.
 * External References: none
 * Returned values: none
 * Errors detected: this is the end of an error chain.
 */
{
  fprintf(stderr, "0x%02X unknown type to display\n", __descriptor(g));
  exit(-1);
}



void display(FILE *fh, Generic *g)
/*     Display generic values by running them through the table.
 * Parameters:
 *   fh: File to write to.
 *   g: Generic value to display.
 * External References:
 *   displayFns[]: Table by type descriptor.
 *   MAXTYPES: size of table.
 * Returned values: none
 * Errors Detected: illegal type descriptor causes halt.
 */
{
  if (g == NULL)
  {
    fprintf(fh, "NULL");
    return;
  }
  if (__descriptor(g) >= MAXTYPES) illegalType(fh, g);
  (displayFns[__descriptor(g)])(fh, g);
}



void initDisplay(void)
/*     Initialize the display table to all illegal. When new types show up, they should override
 *   the illegal. This must be called before any descriptors do their fill in.
 * Parameters: none
 * External References:
 *   displayFns[]: Table by type descriptor.
 *   MAXTYPES: size of table.
 * Returned values: none
 * Errors Detected: none
 */
{
  int32_t i;
  for (i = 0; i < MAXTYPES; ++i) displayFns[i] = illegalType;
}


// EOF
