/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: equal.c
 * Author: Jed Marti
 * Description: Table driven equality for dynamic structures.
 * Created: Sun Dec 15 06:47:04 2024
 * Revision History:
 *   Sun Dec 29 10:58:21 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>

#include "protobuf.h"

// A table of the types. If you add a new type you must add its equality checking function to
// this table. Such a function must have two Generic type arguments. They are both guaranteed
// to be your new type. You should return 1 if the two are equal and 0 if not. The function
// should be added to this table during initialization. We do a descriptor check before the
// call so there's no need for an illegal type.
equalFn equalFns[MAXTYPES];



uint8_t equal(Generic *a, Generic *b)
/*     Compare dynamic structures to the lowest level for equality. Table driven at atoms.
 * Parameters:
 *   a, b: Two Generics to compare (can be NULL).
 * External References:
 *   MAXTYPES: constant, how many types are tghere.
 * Returned values: 1-values equal, 0 not.
 * Errors detected: illegal type descriptor on entity.
 */
{
  // Degenerate cases.
  if (a == NULL)
    if (b == NULL) return 1;
    else return 0;
  else if (b == NULL) return 0;

  // Both have descriptors. Check validity.
  if (__descriptor(a) >= MAXTYPES || __descriptor(b) >= MAXTYPES)
  {
    fprintf(stderr, "%s:%s:%d equal 0x%02X 0x%02X unknown descriptors", __FILE__, __func__,
            __LINE__, __descriptor(a), __descriptor(b));
    exit(-1);
  }

  // Wrong types are false (i.e. 0 != 0.0)
  if (__descriptor(a) != __descriptor(b)) return 0;

  // The have the same descriptor - call the checker for each type.
  return (equalFns[__descriptor(a)])(a, b);
}



// EOF
