/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: listen.c
 * Author: Jed Marti
 * Description: Listen for a connection on the port number given. No timeout.
 * Created: Mon Dec  2 06:15:32 2024
 * Revision History:
 *   Sun Dec 29 14:24:56 2024 - code review
 */

#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netdb.h>
#include <netinet/in.h>


// Global file handle. If -1, then not assigned yet.
int32_t FD = -1;
int32_t SOCKET;

// Noise control.
extern uint8_t Noise;


void netListen(int32_t portnum)
/*     Listen on a port forever.
 * Parameters:
 *   portnum: 1024+ are acceptable, below that you must be super user.
 * External References:
 *   FD: file handle for I/O. -1 means to net connection.
 *   SOCKET: Socket for restart after losing a connections.
 *   Noise: How much noise to generate about what's happening.
 * Returned values: none
 * Errors detected: no available sockets, other OS errors.
 */
{
  static struct sockaddr_in my_addr;
  static struct sockaddr_in his_addr;
  int32_t on;
  int32_t from_len;
  uint32_t hisipv4;

  // Noise maker
  if (Noise > 0) printf("%s:%s:%d listening on port %d\n", __FILE__, __func__, __LINE__, portnum);

  // Get a socket 
  SOCKET = socket(AF_INET, SOCK_STREAM, 0);
  if (SOCKET < 0)
  {
    fprintf(stderr, "%s:%s:%d No available sockets because %sd\n", __FILE__, __func__, __LINE__,
            strerror(errno));
    exit(-1);
  }
  if (Noise > 4) printf("%s:%s:%d Using socket %d\n", __FILE__, __func__, __LINE__, SOCKET);

  // Setup portnum`
  memset(&my_addr, 0, sizeof(my_addr));
  my_addr.sin_family = AF_INET;
  my_addr.sin_addr.s_addr = INADDR_ANY;
  my_addr.sin_port = htons(portnum);
    
  // Get the port
  if (setsockopt(SOCKET, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)))
  {
    fprintf(stderr, "%s:%s:%d Can't reuse socket on port %d because %s\n", __FILE__, __func__,
            __LINE__, portnum, strerror(errno));
    exit(-1);
  }
  if (Noise > 4) printf("%s:%s:%d setsockopt ok\n", __FILE__, __func__, __LINE__);
  
  if (bind(SOCKET, (struct sockaddr *) &my_addr, sizeof(my_addr)) < 0 )
  {
    fprintf(stderr, "%s:%s:%d Can't bind to port number %d on this socket because %s\n", __FILE__,
            __func__, __LINE__, portnum, strerror(errno));
    exit(-1);
  }
  if (Noise > 4) printf("%s:%s:%d bind success\n", __FILE__, __func__, __LINE__);

  //  Do the listen
  if (listen(SOCKET, 1) < 0)
  {
    fprintf(stderr, "%s:%s:%d Listen fails because %s\n", __FILE__, __func__, __LINE__,
            strerror(errno));
    exit(-1);
  }
  from_len = sizeof(his_addr);
  if (Noise > 4) printf("%s:%s:%d listen len=%d\n", __FILE__, __func__, __LINE__, from_len);

  FD = accept(SOCKET, (struct sockaddr *) &his_addr, (uint32_t *) &from_len);
  
  hisipv4 = his_addr.sin_addr.s_addr;
  if (Noise > 4) printf("%s:%s:%d accept of %d.%d.%d.%d finishes with FD=%d\n", __FILE__,
                        __func__, __LINE__, hisipv4 >> 24, (hisipv4 >> 16) & 0xFF,
                        (hisipv4 >> 8) & 0xFF, hisipv4 & 0xFF, FD);
  if (FD < 0)
  {
    fprintf(stderr, "%s:%s:%d Listen fails because %s\n", __FILE__, __func__, __LINE__,
            strerror(errno));
    exit(-1);
  }

  // Close the socket, we have the FD. Leave it open so we can reuse?
  //  close(SOCKET);
  if (Noise > 0) printf("%s:%s:%d connection handle %d\n", __FILE__, __func__, __LINE__, FD);
}



void netAccept(void)
/*     Once we close the connection to a Meshtastic device, we can wait for a new one without
 *   going through the whole socket regimaroles.
 * Parameters: none
 * External References:
 *   SOCKET: socket activated at startup.
 *   FD: File handle when connection opens.
 *   Noise: how muich noise to generate about what we're doing.
 * Returned values: none
 * Errors detected: listen fails, couldn't open file.
 */
{
  static struct sockaddr_in his_addr;
  int32_t from_len;

  // Wait for a remote.
  if (listen(SOCKET, 1) < 0)
  {
    fprintf(stderr, "%s:%s:%d Listen fails because %s\n", __FILE__, __func__, __LINE__,
            strerror(errno));
    exit(-1);
  }

  // Display his address and finish the connection.
  from_len = sizeof(his_addr);
  if (Noise > 5) printf("%s:%s:%d listen len=%d\n", __FILE__, __func__, __LINE__, from_len);

  // Get the file handle.
  FD = accept(SOCKET, (struct sockaddr *) &his_addr, (uint32_t *) &from_len);
  if (Noise > 5) printf("%s:%s:%d accept finishes with FD=%d\n", __FILE__, __func__, __LINE__, FD);
  if (FD < 0)
  {
    fprintf(stderr, "%s:%s:%d Listen fails because %s\n", __FILE__, __func__, __LINE__,
            strerror(errno));
    exit(-1);
  }
}



// EOF
