/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: mqtt.c
 * Author: Jed Marti
 * Description: Control program for MQTT servicing.
 * Created: Mon Dec  2 06:21:10 2024
 * Revision History:
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <setjmp.h>
#include <signal.h>
#include <unistd.h>
#include <sys/socket.h>

#include "mqtt.h"
#include "protobuf.h"

// See listen.c. -1 until we have a net connection.
extern int32_t FD;
extern int32_t SOCKET;

// Port number.
extern int32_t portNumber;

// Abbreviation file name.
extern char *abbreviationFileName;

// Nose level.
extern uint8_t Noise;

// Downlink status.
uint8_t downStatus = downNoConnection;

// If remote dies.
jmp_buf keepAliveNot;
extern uint32_t keepAlive;
extern uint32_t keepAliveStartup;

// Log files, see log.c, used by LOG macro
extern char logMsg[];
extern FILE *LOGfh;

// Testing.
extern uint8_t T[];
extern uint16_t TINDEX;
extern protoBlock protoServiceEnvelope;

// See io.c
extern uint8_t isStarted;
extern uint8_t startSTDIN;



void noPINGREQ(int32_t v)
/*     A CONNECT mentioned that it will do a keep alive and gives the number of sections that
 *   are stored in keepAlive. This watchdog alarm response gets reset every time we get such
 *   a message so it doesn't trigger the disconnect.
 * Parameters:
 *   v: set by the longjmp that got us here. Ignored.
 * External References:
 *   keepAliveNot: where to go to close the connection.
 * Returned values: none.
 * Errors detected: none here.
 */
{
  LOG("No keep alive ping request, closing network");
  if (Noise > 0)
    printf("%s:%s:%d no keep alive, resetting\n", __FILE__, __func__, __LINE__);
  longjmp(keepAliveNot, 1);
}



void sigIntExit(int32_t x)
/*     ^C (well two of them) means the user wants termination. Log this, close the log, and
 *   exit. This is the result of a SIGINT alarm.
 * Parameters:
 *   x: ignored.
 * External References:
 *   LOGfh: Log file handle.
 * Returned values: none.
 * Errors detected: Log stuff may have an error.
 */
{
  LOG("Terminated by ^C");
  fclose(LOGfh);
  fprintf(stderr, "Terminated by ^C\n");
  exit(-1);
}


void sigTermExit(int32_t x)
/*     An error exit causes a log entry and termination.
 * Parameters:
 *   x: ignored.
 * External References:
 *   LOGfh: Log file handle.
 * Returned values: none
 * Errors detected: none
 */
{
  LOG("Terminated");
  fclose(LOGfh);
  fprintf(stderr, "Terminated\n");
  exit(-1);
}



int32_t main(int32_t argc, char *argv[])
/*     Main program for MQTT service.
 * Parameters: typical command line stuff.
 * External References:
 *   abbreviationFileName: abbreviations for mail service.
 *   portNumber: where to listen for MQTT messages.
 *   FD: file handle for reading/writing TCP connection.
 *   Noise: how much noise to generate: 0=none 5=lots.
 * Returned values: exits give return code of -1 or 0 for normal termination.
 * Errors detected: none here.
 */
{
  int32_t pnum;
  
  printf("mqtt sms server V%d.%02d.%03d\n", VERSION, RELEASE, PATCH);

  // Parse command line options.
  cmdline(argc, argv);

  // Initialize LOG.
  initLog();
  LOG("Startup mqtt V%d.%02d.%03d", VERSION, RELEASE, PATCH);

  // Check gateways. If more than one, then we have a problem.
  checkGateways();
  
  // Set up signals.
  signal(SIGINT, sigIntExit);
  signal(SIGTERM, sigTermExit);

  // Set up for protobufs. Don't futz with the order.
  initProtobuf();
  initGC();
  initDisplay();
  initClone();
  
  initScalars();
  initStrings();
  initProtoBufs();

  //  text2protoBuf("Test #4");
  //  mqtt2Mesh(T, TINDEX);
  //  display(stdout, protoBuf2Generic(T, TINDEX, &protoServiceEnvelope));
  //  printf("\n");
  //  exit(0);

  // Read abbreviations and start TCP (required by Windows).
  initAbbreviations(abbreviationFileName);

   // Wait for local device on default or specified port.
  netListen(portNumber);

  // We lost the remote, wait for a new to take over.
 restart:
  if (Noise > 1) printf("Opened device on %d on port %d\n", FD, portNumber);
  LOG("Opened Meshtastic on port %d", portNumber);

  // Set alarm for if we don't get our ping at keepAliveStartup more seconds than the keepAlive
  // value. If the alarm goes off, then we close the socket and wait for a new connection.
  signal(SIGALRM, noPINGREQ);
  alarm(keepAlive + keepAliveStartup);
  if (setjmp(keepAliveNot) != 0)
  {
    close(FD);
    LOG("Meshtastic connection lost after %d seconds, restarting", keepAlive + keepAliveStartup);
    netAccept();
    goto restart;
  }
  
 loop:
  parseMQTTPacket();

  if (FD == -1)
  {
    LOG("Meshtastic connection lost, restarting");
    netAccept();
    goto restart;
  }

  // For stdout tracing.
  fflush(stdout);

  goto loop;
}



// EOF
