/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: pingreq.c
 * Author: Jed Marti
 * Description: Respond to a ping request - we are here! Reset watchdog in mqtt main so we don't
 *   disconnect prematurely.
 * Created: Mon Dec  2 06:40:00 2024
 * Revision History:
 *   Sun Dec 29 17:29:09 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

#include "mqtt.h"

// The keep alive time.
extern uint32_t keepAlive;

// File handle for net.
extern int32_t FD;

// Noise level.
extern uint8_t Noise;

// Log files, see log.c, used by LOG macro.
extern char logMsg[];
extern FILE *LOGfh;

// Response to ping request is fixed - two bytes to write.
static uint8_t prq[] = {MQTT_PINGRESP << 4, 0 };


void mqttPingReq(uint8_t pt, uint8_t *pkt)
/*     A ping request from the Meshtastic device. Respond with a PINGRESP packet. If we don't
 *   reset the alarm, an interrupt occurs (see mqtt.c) that shuts down the network connection
 *   and relistens for another.
 * Parameters:
 *   pt: Type (PINGREQ).
 *   pkt: Address of packet data (ignored).
 * External References:
 *   Noise: How much noise to make.
 *   FD: TCP file handle for writing response.
 * Returned values: none
 * Errors detected: Write PINGRESP fails, causes halt.
 */ 
{
  ssize_t n;

  // Clear the alarm and set it for the next time interval. The alarm only goes off if we don't
  // get this message every keepAlive seconds + some noise slop (10 seconds).
  alarm(keepAlive + 10);
  
  if (Noise > 0) printf("PINGREQ\n");
  if (write(FD, prq, 2) == -1)
  {
    LOG("PINGRESP failed because %s",  strerror(errno));
    fprintf(stderr, "%s:%s:%d PINGRESP failed because %s\n", __FILE__, __func__, __LINE__,
            strerror(errno));
    exit(-1);
  }
}


// EOF
