/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: position.c
 * Author: Jed Marti
 * Description: Keep track of positions we've seen.
 * Created: Sat Dec 14 08:29:26 2024
 * Revision History:
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "mqtt.h"
#include "protobuf.h"

// Noise level.
extern uint8_t Noise;

// The master protoBuf description for a position.
extern protoBlock protoPosition;

// A an association list of (MAC . ...). The alist value is whatever showed up in the packet.
// When we get a message from a remote we look for it here to get its last known position that
// we can tack onto the message being forwarded.
Generic *REMOTES = NULL;


Generic *lookFor(int64_t num, Generic *l)
/*     Check the list of remotes for the station MAC address and return its entry in the
 *   list. 
 * Parameters:
 *   num: The last 4 bytes of the MAC address.
 *   l: The association list of ((remote_mac . <position>) ...)
 * External References: none
 * Returned Values: The associated list (usually an a-list) or NULL if not found.
 * Errors detected: none
 */
{
  for (; __isPair(l); l = __cdr(l))
  {
    if (__int(__caar(l)) == num) return __cdar(l);
  }
  return NULL;
}



void position(Generic *g)
/*     We have a Position or PositionLite packet. Pick it up and add to the list. Note that the
 *   PositionLite packet is a subset of Position.
 * Parameters:
 *   p: The MQTT packet as parsed. Now we do protoBuf.
 * External References:
 *   Noise: for tracing.
 */
{
  Generic *pkt, *data, *payload, *from;
  Generic *prev, *pdat;

  if (Noise > 4) printf("%s:%s:%d Starting position\n", __FILE__, __func__, __LINE__);  
  pkt = assoc("packet", g);
  if (pkt == NULL)
  {
    fprintf(stderr, "%s:%s:%d Service envelope packet missing Data\n", __FILE__, __func__,
            __LINE__);
    return;
  }
  // MAC addr used as station id in our list. We get a 32 bit number out of this.
  from = assoc("from", pkt);
  if (from == NULL)
  {
    fprintf(stderr, "%s:%s:%d MeshPacket missing from\n", __FILE__, __func__, __LINE__);
    return;
  }
    
  data = assoc("data", pkt);
  if (data == NULL)
  {
    fprintf(stderr, "%s:%s:%d Packet missing Data\n", __FILE__, __func__,
            __LINE__);
    return;
  }

  payload = assoc("payload", data);
  if (payload == NULL)
  {
    fprintf(stderr, "%s:%s:%d Data missing payload\n", __FILE__, __func__,
            __LINE__);
    return;
  }

  // Parse data. If it's a ProtoLite packet, the search will just terminate early.
  pdat = protoBuf2Generic(__string(payload), __strlen(payload), &protoPosition);
  if (Noise > 5)
  {
    printf("%s:%s:%d Position = ", __FILE__, __func__, __LINE__);
    display(stdout, pdat);
    printf("\n");
  }

  // See if this remote is already on the list.
  if ((prev = lookFor(__int(from), REMOTES)) != NULL)
  {
    // Clear the old stuff.
    printf("Freeing old remote\n");
    gc(__cdr(prev));
    printf("done\n");
    __cdr(prev) = pdat;
  }

  // It's not so add it.
  else REMOTES = cons(cons(from, pdat), REMOTES);
  return;
}


// EOF
