/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: protobuf.c
 * Author: Jed Marti
 * Description: Read protobuf defined stuff and return JSON like structures.
 * Created: Fri Dec  6 08:36:05 2024
 * Revision History:
 *   Sun Dec 29 19:39:44 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "mqtt.h"
#include "protobuf.h"

// See mprotos.c. These are the protobuf messages we're being asked to decode starting with
// protoServiceEnvelope.
extern protoBlock *subBlocks[];
extern protoBlock protoServiceEnvelope;
extern protoBlock protoUser;

// General noise making, 0=none.
extern uint8_t    Noise;

// Log files, see log.c, used by LOG macro
extern char logMsg[];
extern FILE *LOGfh;

// The buffer to read and convert into Generic (like JSON) structures. We copy stuff here to
// avoid messing with the MQTT copy.
uint8_t IBUF[1024];
int16_t INDEX;
int16_t LAST;

// The table of conversion functions from the primitives.
protoread convertProtos[MAXPROTOS];


static uint8_t pgetc(void)
/*     Get the next character from the input buffer.
 * Parameters: none
 * External References:
 *   IBUF[]: Our copy of the message payload.
 *   INDEX: Index to the character to return.
 *   LAST: Error if we get to this point.
 * Returned values: a byte.
 * Errors detected: Falling off the end causes a halt.
 */
{
  if (INDEX >= LAST)
  {
    fprintf(stderr, "%s:%s:%d Premature payload end, quitting\n", __FILE__, __func__, __LINE__);
    LOG("Premature payload end");
    exit(-1);
  }
  return IBUF[INDEX++];
}



uint64_t readVARINT(void)
/*     Read a variable integer, one of the protoBuf types, and return as 64 bits. Given the
 *   Meshtastic message size seeing more than 2 bytes is probably an error.
 * Parameters: none
 * External References: none
 * Returned values: a 64 bit int using 7 bits of input as per protoBuf definition. A rather
 *   clumsy implementation.
 * Errors detected: none here.
 */
{
  uint64_t v;
  uint8_t  b;

  b = pgetc();
  if (b & 0x80)
  {
    // At least 16 bits here.
    v = (uint64_t) (b & 0x7F);
    b = pgetc();
    if (b & 0x80)
    {
      // At least 24 bits here.
      v |= ((uint64_t) (b & 0x7F)) << 7;
      b = pgetc();
      if (b & 0x80)
      {
        v |= ((uint64_t) (b & 0x7F)) << 14;
        b = pgetc();
        if (b & 0x80)
        {
          v |= ((uint64_t) (b & 0x7F)) << 21;
          b = pgetc();
          if (b & 0x80)
          {
            v |= ((uint64_t) (b & 0x7F)) << 28;
            b = pgetc();
            if (b & 0x80)
            {
              v |= ((uint64_t) (b & 0x7F)) << 35;
              b = pgetc();
              if (b & 0x80)
              {
                v |= ((uint64_t) (b & 0x7F)) << 42;
                b = pgetc();
                if (b & 0x80)
                {
                  v |= ((uint64_t) (b & 0x7F)) << 49;
                  b = pgetc();
                  if (b & 0x80)
                  {
                    v |= ((uint64_t) (b & 0x7F)) << 56;
                    b = pgetc();
                    if (b & 0x80)
                    {
                      v |= ((uint64_t) (b & 0x7F)) << 63;
                      return v;
                    }
                    return (((uint64_t) b) << 63) | v;
                  }
                  return (((uint64_t) b) << 56) | v;

                }
                return (((uint64_t) b) << 49) | v;
              }

              // 56 bit unsigned.
              return (((uint64_t) b) << 42) | v;
            }

            // 48 bit unsigned.
            return (((uint64_t) b) << 35) | v;
          }

          // 40 bit unsigned.
          return (((uint64_t) b) << 28) | v;
        }

        // 32 bit unsigned.
        return (((uint64_t) b) << 21) | v;
      }

      // 24 bit unsigned.
      return (((uint64_t) b) << 14) | v;
    }

    // 16 bit unsigned.
    return (((uint64_t) b) << 7) | v;
  }

  return (uint64_t) (b & 0x7F);
}



static uint32_t readI32(void)
/*     Read a 32 bit value as 4 bytes and return. One of the protoBuf primitive types.
 * Parameters: none
 * External References:
 *    Noise: Trace output value if > 4.
 * Returned values: 4 bytes little endian (?)
 * Errors detected: none here.
 */
{
  uint32_t v;
  v = pgetc();
  v |= ((uint32_t) pgetc()) << 8;
  v |= ((uint32_t) pgetc()) << 16;
  v |= ((uint32_t) pgetc()) << 24;
  if (Noise > 4) printf("%s:%s:%d I32 value 0x%08x %u\n", __FILE__, __func__, __LINE__, v, v);
  return v;
}



static uint64_t readI64(void)
/*     Read a 64 bit value as 8 bytes and return. One of the protoBuf primitive types.
 * Parameters: none
 * External References:
 *    Noise: Trace output value if > 4.
 * Returned values: 8 bytes little endian (?)
 * Errors detected: none here.
 */
{
  uint64_t v;

  v = pgetc();
  v |= ((uint64_t) pgetc()) << 8;
  v |= ((uint64_t) pgetc()) << 16;
  v |= ((uint64_t) pgetc()) << 24;
  v |= ((uint64_t) pgetc()) << 32;
  v |= ((uint64_t) pgetc()) << 40;
  v |= ((uint64_t) pgetc()) << 48;
  v |= ((uint64_t) pgetc()) << 56;
  if (Noise > 4) printf("%s:%s:%d I64 value 0x%016lX %ld\n", __FILE__, __func__, __LINE__, v, v);
  return v;
}
  


static Generic *protoRead1(protoBuf *p)
/*     Convert one of the raw values into a Generic by calling the appropriate conversion routine.n
 * Parameters:
 *   p: A protoBuf entry describing what to look for.
 * External References:
 *   convertProtos[]: A vector of function calls to do the conversion. Set up during 
 *     initialization.
 * Returned values: a Generic with the type filled in.
 * Errors detected: none, but should probably check ptype range.
 */
{
  if (p->ptype >= MAXPROTOS)
  {
    fprintf(stderr, "%s:%s:%d Target index %d illegal (max is %d)\n", __FILE__, __func__, __LINE__,
            p->ptype, MAXPROTOS);
    exit(-1);
  }
  return (convertProtos[p->ptype])();
}



static Generic *protoInt32(void)
/*     Convert the 64 bit unsigned into a signed 32 bitter stored as a 64 bit value. The input
 *   is a VARINT.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  return newInt((int64_t) (int32_t) readVARINT());
}
 


static Generic *protoInt64(void)
/*     Convert the 64 bit unsigned into a signed 64 bitter stored as a 64 bit value. The input
 *   is a VARINT.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  return newInt((int64_t) readVARINT());
}



static Generic *protoUint32(void)
/*     Convert the 64 bit unsigned into a signed 32 bitter stored as a 64 bit value so it will 
 *   always. The input is a VARINT.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  return newInt((uint32_t) readVARINT());
}



static Generic *protoUint64(void)
/*     Just pack an unsigned 64 bitter into a generic signed 64 bitter. The input is a VARINT.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  return newInt(readVARINT());
}



static Generic *protoSint32(void)
/*     Convert the 64 bit unsigned into a signed 32 bitter stored as a 64 bit value. Sign
 *   extend the result back into 64 bits. The input is a VARINT.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  uint32_t r;
  r = (uint32_t) readVARINT();
  
  if (r & 1) return newInt(- ((r >> 1) + 1));
  else return newInt(r >> 1);
}



static Generic *protoSint64(void)
/*     Convert the 64 bit unsigned into a signed 64 bitter stored as a 64 bit value. The input is
 *    a VARINT.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  uint64_t r;
  r = readVARINT();
  if (r & 1) return newInt(- ((r >> 1) + 1));
  else return newInt(r >> 1);
}



static Generic *protoBool(void)
/*     Convert a VARINT into 0/1 stored as a Generic.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  return newInt(readVARINT() & 0x00001);
}



static Generic *protoEnum(void)
/*     Convert a VARINT into an integer. These are usually small.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  return newInt(readVARINT());
}



static Generic *protoFixed64(void)
/*     Read the 8 bytes and return as a Generic Int. In some cases this will be smaller than a
 *   VARINT. It's unsigned but stored as a signed. Get over it.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  return newInt(readI64());
}



static Generic *protoSFixed64(void)
/*     Read the 8 bytes and return as a Generic Int. In some cases this will be smaller than a
 *   VARINT.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: none
 */
{
  return newInt(readI64());
}



static Generic *protoFixed32(void)
/*     Read a 32 bit unsigned integer, convert to 64 bit (no sign extend).
 * Parameters: none
 * External References:
 *   Noise: Generate trace info if > 4.
 * Returned values: Generic 64 bit version of number.
 * Errors detected: none here.
 */
{
  uint64_t i;
  i = (uint64_t) readI32();
  if (Noise > 4) printf("%s:%s:%d i=0x%0lX\n", __FILE__, __func__, __LINE__, i);
  return newInt(i);
}



static Generic *protoSFixed32(void)
/*     Read a 32 bit unsigned integer, convert to 64 bit sign extended.
 * Parameters: none
 * External References:
 *   Noise: Generate trace info if > 4.
 * Returned values: Generic 64 bit version of number.
 * Errors detected: none here.
 */
{
  int64_t v;

  v = readI32();
  if (v & 0x80000000) v |= 0xFFFFFFFF80000000;
  if (Noise > 4) printf("%s:%s:%d i=0x%ld\n", __FILE__, __func__, __LINE__, v);
  return newInt((int64_t) v);
}



static Generic *protoDouble(void)
/*     Read a 64 bit double-precision floating-point value.
 * Parameters: none
 * External References:
 *   Noise: Generate trace info if > 4.
 * Returned values: Generic Double.
 * Errors detected: none here.
 */
{
  uint64_t v;
  double   *d;
  v = readI64();
  d = (double *) &v;
  if (Noise > 4) printf("%s:%s:%d d=%g\n", __FILE__, __func__, __LINE__, *d);
  return newDouble(*d);
}



static Generic *protoFloat(void)
/*     Read a 32 bit single-precision floating-point value.
 * Parameters: none
 * External References:
 *   Noise: Generate trace info if > 4.
 * Returned values: Generic Double.
 * Errors detected: none here.
 */
{
  uint32_t v;
  float    *f;
  v = readI32();
  f = (float*) &v;
  if (Noise > 4) printf("%s:%s:%d f=%g\n", __FILE__, __func__, __LINE__, *f);
  return newDouble((float) *f);
}



uint8_t *readLEN(uint32_t *sz)
/*     Read a string with a VARINT length and return as a plain string. This could be ASCII
 *   characters but also includes 'sub records' that will be any sort of values.
 * Parameters: none
 * External References:
 *   Noise: >4, show the size.
 * Returned values: A byte vector.
 * Errors detected: Out of space allocating the string causes halt.
 */
{
  uint32_t i;
  uint8_t  *str;

  *sz = readVARINT();
  if (Noise > 4) printf("%s:%s:%d sz = %u\n", __FILE__, __func__, __LINE__, *sz);
  if ((*sz + INDEX) > LAST)
  {
    printf("%s:%s:%d Size %u runs off end %d -> %d\n", __FILE__, __func__, __LINE__, *sz,
           INDEX, LAST);
    INDEX += *sz;
    return NULL;
  }

  // Get enough space.
  if ((str = (uint8_t *) malloc(*sz + 1)) == NULL) outOfSpace("LEN string");

  // Copy string to our new buffer. Sometimes it's a real string so add a 0 just for grins.
  for (i = 0; i < *sz; ++i) str[i] = pgetc();
  str[i] = 0;
  return str;
}



static Generic *protoLen(void)
/*     Here we've been requested to get a string from a protobuf description, not a sub record.
 *   Read the string and return as a generic.
 * Parameters: none
 * External References: none
 * Returned values: A generic signed integer.
 * Errors detected: Out of space allocating the string causes halt.
 */
{
  uint8_t *str;
  uint32_t sz;
  str = readLEN(&sz);
  if (Noise > 4) printf("%s:%s:%d string '%s'[%u]\n", __FILE__, __func__, __LINE__, str, sz);
  return newString(str, sz);
}



// Forward reference.
static Generic *protoBuf2Generic1(protoBlock *bl);

static Generic *protoSubPacket(uint8_t sub)
/*     We have a request for a subpacket whose number we know from the protobuf description,
 *   not the message.
 * Parameters:
 *   bl: The proto block describing this sub packet.
 * External References:
 *   Noise: how much trace to generate.
 *   INDEX, LAST: into our copy of the message.
 *   IBUF: copy of the input buffer.
 * Returned values: A Generic structure for this thing.
 * Errors detected: illegal substructure number, no sub block
 */
{
  uint32_t   sz;
  uint32_t   oldLAST, startPos;
  protoBlock *nbl;
  Generic    *g;

  // Trace and error check.
  if (Noise > 3) printf("%s:%s:%d subgroup block %d\n", __FILE__, __func__, __LINE__, sub);
  if (sub >= MAXSUBS)
  {
    fprintf(stderr, "%s:%s:%d Sub record number %u illegal\n", __FILE__, __func__, __LINE__, sub);
    exit(-1);
  }
  
  // We have a subgroup. Pick up the new protoBlock or error if NULL.
  nbl = subBlocks[sub];
  if (nbl == NULL)
  {
    fprintf(stderr, "%s:%s:%d Bad subgroup %d\n", __FILE__, __func__, __LINE__, sub);
    exit(-1);
  }
  
  // Save where we are before we get the string.
  oldLAST = LAST;
  
  // The number of bytes in the packet, set the new end.
  sz = readVARINT();
  startPos = INDEX;
  LAST = INDEX + sz;

  if (Noise > 4)
    printf("%s:%s:%d Subgroup size %d, LAST now %d\n", __FILE__, __func__, __LINE__, sz, LAST);

  // Now get the new buffer, trace the result. Here we're passing along the new control
  // block.f
  printf("0x%02X 0x%02X\n", IBUF[INDEX], IBUF[INDEX + 1]);
  g = protoBuf2Generic1(nbl);
  if (Noise > 4)
  {
    printf("%s:%s:%d ", __FILE__, __func__, __LINE__);
    display(stdout, g);
    printf("\n");
  }
  
  // Restore LAST so we can continue with earlier packet. INDEX should now point to the
  // correct byte.
  LAST = oldLAST;

  return g;
}
  
 
  

static protoBuf *index2protoBuf(int32_t indx, protoBuf *p)
/*     Search A protoBuf description for an index and return the description if available.
 * Parameters:
 *   indx: Index to search for.
 *   p: A vector of protoBuf descriptions.
 * External References: none
 * Returned values: The matching protoBuf or NULL if not found.
 * Errors detected: none
 */
{
  int32_t i;
  if (Noise > 4) printf("%s:%s:%d look for index %d\n", __FILE__, __func__, __LINE__, indx);
  for (i = 0; p[i].index != -1; ++i)
    if (p[i].index == indx)
    {
      if (Noise > 4)
        printf("%s:%s:%d Found index %d name %s\n", __FILE__, __func__, __LINE__, indx, p[i].name);
      return &p[i];
    }
  return NULL;
}



static Generic *protoBuf2Gen(protoBuf *p, protoBlock *bl)
/*     We're working with a particular field. Find the support function to read that and
 *   return a pair with the name and value.
 * Parameters:
 *   p: A protoBuf structure describing this field.
 *   bl: The protoBlock structure containing this field.
 * External References: none except constants.
 * Returned Values: A Generic structure.
 */
{
 // The wire type is hidden in the generic type except for subblocks.
  if (Noise > 4)
    printf("%s:%s:%d INDEX=%d, LAST=%d name=%s gtype=%d ptype=%d sub=%d\n",
           __FILE__, __func__, __LINE__, INDEX, LAST, p->name, p->gtype, p->ptype, p->sub);

  // Subgroups done differently.
  if (p->gtype == SGROUP)
  {
    if (Noise > 4) printf("%s:%s:%d working subgroup %s\n", __FILE__, __func__, __LINE__, p->name);
    return cons(newString(p->name, strlen(p->name)), protoSubPacket(p->sub));
  }

  // Others go through the table.
  else
  {
    //    printf("0x%0lX\n", (uint64_t) convertProtos[p->ptype]);
    return cons(newString(p->name, strlen(p->name)), (convertProtos[p->ptype])() );
  }
}
    


static Generic *protoBuf2Generic2(protoBlock *bl)
/*     We're at the start of an entry. Just do it and return.
 * Parameters:
 *   bl: The control block (description).
 * External References:
 *   LAST: The size of the message in bytes.
 *   INDEX: a global pointer.
 *   Noise: Say that we're starting if noisy.
 * Returned values: The key value pair (alist).
 * Errors detected: Protobuf index not found.
 */
{
  uint8_t q;
  uint8_t wire;
  protoBuf *pb;
  Generic  *g;

  // Get the descriptor and index. This is a VARINT otherwise 

  q = pgetc();
  wire = q & 0x07;

  if (Noise > 0)
    printf("%s:%s:%d q=0x%0X wire type %d, index %d block name %s\n", __FILE__, __func__, __LINE__,
           q, q & 0x07, q >> 3, bl->name);

  // Position message seems to have this problem.
  if (q == 0) return NULL;
  
  // TLV is a VARINT so remove top bit. Is this true?
  pb = index2protoBuf(q >> 3, *(bl->p));
  if (pb == NULL)
  {
    fprintf(stderr, "%s:%s:%d Index %d not found in %s\n", __FILE__, __func__, __LINE__,
            q >> 3, bl->name);
    exit(-1);
  }
  if (Noise > 2) printf("%s:%s:%d %s sub=%d\n", __FILE__, __func__, __LINE__, pb->name, pb->sub);

  // Now do the single conversion.
  g = protoBuf2Gen(pb, bl);
  if (Noise > 2)
  {
    printf("%s:%s:%d Result is ", __FILE__, __func__, __LINE__); fflush(stdout);
    display(stdout, g);
    printf("\n");
  }
  return g;
}



static Generic *protoBuf2Generic1(protoBlock *bl)
/*     Do an entire block (this could be a subblock).
 * Parameters:
 *   bl: The control block (description).
 * External References:
 *   LAST: The size of the message in bytes.
 *   INDEX: a global pointer.
 *   Noise: Say that we're starting if noisy.
 * Returned values: The key value pair (alist).
 * Errors detected: none here.
 */
{
  Generic *l = NULL, *le;
  Generic *g;

  while (INDEX < LAST)
  {

    // Last 2 are bizarre.
    if (INDEX >= (LAST - 2))
    {
      if (Noise > 4)
        printf("%s:%s:%d Ignoring last 2 bytes 0x%02X, 0x%02X\n", __FILE__, __func__, __LINE__,
               IBUF[INDEX], IBUF[INDEX + 1]);
      INDEX = LAST;
      break;
    }
    if (Noise > 4)
      printf("%s:%s:%d Block %s INDEX=%d, LAST=%d\n", __FILE__, __func__, __LINE__,
             bl->name, INDEX, LAST);
    g = protoBuf2Generic2(bl);
    
    if (l == NULL) l = le = cons(g, NULL);
    else
    {
      __cdr(le) = cons(g, NULL);
      le = __cdr(le);
    }
    if (Noise > 2)
    {
      printf("%s:%s:%d Partial list ", __FILE__, __func__, __LINE__);
      display(stdout, l);
      printf("\n");
    }
  }
  return l;
}



Generic *protoBuf2Generic(uint8_t *ibuf, int16_t plen, protoBlock *bl)
/*     Convert a buffer to a key-value list and return.
 * Parameters:
 *   ibuf[]: A payload from MQTT.
 *   int16_t: The number of bytes there.
 *   p[]: A vector of field descriptors.
 *   name: The block name.,
 * External References:
 *   IBUF[]: Copy the message payload here.
 *   LAST: The size of the message in bytes.
 *   INDEX: a global pointer.
 *   Noise: Say that we're starting if noisy.
 * Returned values: The key value pair (alist).
 * Errors detected: none here.
 */
{
  // Get stuff to our workspace.
  memcpy(IBUF, ibuf, plen),
    LAST = plen,
    INDEX = 0;

  if (Noise > 4) printf("%s:%s:%d starting\n", __FILE__, __func__, __LINE__);
  return protoBuf2Generic1(bl);
}



/*
void dumppb(protoBlock *bl)
// Tracing.
{
  int32_t i;
  protoBuf *p;
  printf("Name=%s\n", bl->name);
  p = *(bl->p);
  for (i = 0; p[i].index != -1; ++i) printf("%d: %s\n", i, p[i].name);
}
*/


void initProtobuf(void)
/*     Fill in a table that can't be done statically. Must be called at startup.
 * Parameters: none
 * External References:
 *   convertProtos[]: Table of value types.
 *   proto_????: enumeration constant taken from protobuf.h
 * Returned values: none
 * Errors detected: none
 */
{
  // Fill in the conversion protoBuf functions.
  convertProtos[proto_int32] = protoInt32,
    convertProtos[proto_int64] = protoInt64,
    convertProtos[proto_uint32] = protoUint32,
    convertProtos[proto_uint64] = protoUint64,
    convertProtos[proto_sint32] = protoSint32,
    convertProtos[proto_sint64] = protoSint64,
    convertProtos[proto_bool] = protoBool,
    convertProtos[proto_enum] = protoEnum,
    convertProtos[proto_fixed64] = protoFixed64,
    convertProtos[proto_sfixed64] = protoSFixed64,
    convertProtos[proto_double] = protoDouble,
    convertProtos[proto_fixed32] = protoFixed32,
    convertProtos[proto_sfixed32] = protoSFixed32,
    convertProtos[proto_float] = protoFloat,
    convertProtos[proto_len] = protoLen;
}



// EOF
