/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: protobuf.h
 * Author: Jed Marti
 * Description: Prototypes and structures for protobuf.
 * Created: Fri Dec  6 08:36:05 2024
 * Revision History:
 */


#ifndef protobuf_h
#define protobuf_h

// The protobuf primitive types.
enum {
      VARINT =   0,
      I64 =      1,
      LEN =      2,
      SGROUP =   3,
      EGROUP =   4,
      I32 =      5,
      ERROR =    6,
      ERROR2 =   7
      };

// A structure for storing the protobuf descriptions for meshtastic.
typedef struct {
  uint8_t  gtype;              // Use one of the readers above.
  uint8_t  ptype;              // Use one of the converters below.
  uint8_t  sub;                // Use one of the the converters below.
  int16_t  index;              // The field index.
  char     *name;              // A name, mostly for debugging.
} protoBuf;


// The container for each protobuf record.
typedef struct {
  char     *name;              // The protobuf name.
  protoBuf *p[];
} protoBlock;

  
enum {
      proto_int32      = 0,
      proto_int64      = 1,
      proto_uint32     = 2,
      proto_uint64     = 3,
      proto_sint32     = 4,
      proto_sint64     = 5,
      proto_bool       = 6,
      proto_enum       = 7,
      proto_fixed64    = 8,
      proto_sfixed64   = 9,
      proto_double     = 10,
      proto_fixed32    = 11,
      proto_sfixed32   = 12,
      proto_float      = 13,
      proto_len        = 14};
#define MAXPROTOS     15

// These are numbers for subpackets which indicate which protobuf to use. This avoids having
// address in the initialization. See mprotos.c for them.
enum {
      noSub,
      ServiceEnvelopeId,
      MeshPacketId,
      DataId,
      PositionLiteId,
      testSubId,
};
#define MAXSUBS 6

// The portnum enumeration is the Data message type.For any new 'apps'
// that run on the device or via sister apps on phones/PCs they should
// pick and use a unique 'portnum' for their application. If you are
// making a new app using meshtastic, please send in a pull request to
// add your 'portnum' to this master table. PortNums should be
// assigned in the following range: 0-63 Core Meshtastic use, do not
// use for third party apps 64-127 Registered 3rd party apps, send in
// a pull request that adds a new entry to portnums.proto to register
// your application 256-511 Use one of these portnums for your private
// applications that you don't want to register publically All other
// values are reserved. Note: This was formerly a Type enum named
// 'typ' with the same id # We have change to this 'portnum' based
// scheme for specifying app handlers for particular payloads. This
// change is backwards compatible by treating the legacy
// OPAQUE/CLEAR_TEXT values identically.
enum {
      UNKNOWN_AP = 0,
      TEXT_MESSAGE_APP            = 1,
      REMOTE_HARDWARE_APP         = 2,
      POSITION_APP                = 3,
      NODEINFO_APP                = 4,
      ROUTING_APP                 = 5,
      ADMIN_APP                   = 6,
      TEXT_MESSAGE_COMPRESSED_APP = 7,
      WAYPOINT_APP                = 8,
      AUDIO_APP                   = 9,
      DETECTION_SENSOR_APP        = 10,
      REPLY_APP                   = 32,
      PAXCOUNTER_APP              = 64,
      RANGE_TEST_APP              = 66,
      TELEMETRY_APP               = 67,
      ZPS_APP                     = 68,
      SIMULATOR_APP               = 69,
      TRACEROUTE_APP              = 70,
      NEIGHBORINFO_APP            = 71,
      ATAK_PLUGIN                 = 72,
      MAP_REPORT_AP               = 73,
      POWERSTRESS_APP             = 74,
      PRIVATTE_APP                = 256,
      ATAK_FORWARDER_APP          = 257,
      MAX_APP                     = 511
};


// This taken from ja.
typedef struct {
  void         *next; 
  union {
      struct {
        uint8_t size56[7];           // The data payload size in bytes.
        uint8_t descriptor;          // A 7 bit descriptor.
      }  sized;
      struct {
        int32_t v32;                 // For Int32 and symbol. Saves many bytes. 
        uint8_t dummy[3];
        uint8_t descriptor;
      } size32;


      uint64_t  descriptorAndSize;   // Used to get the 64 bit value (free size).
    };
  } Generic;


void proto2list(uint8_t *ibuf);

// Table driven stuff.
typedef void (*gcFn)(Generic *g);
typedef void (*displayFn)(FILE *fh, Generic *g);
typedef Generic *(*cloneFn)(Generic *v);
typedef Generic *(*protoread)(void);
typedef uint8_t (*equalFn)(Generic *a, Generic *b);


// Macros for getting values from the header.
#define __descriptor(x) (((Generic*) (x))->sized.descriptor)
#define __descriptorAndSize(v) (((Generic*) (v))->descriptorAndSize)
#define __next(x) (((Generic*) (x))->next)
#define __size0(x) (((Generic*) (x))->descriptorAndSize)
#define __sizeg(x) ((((Generic*) (x))->descriptorAndSize) & 0x00FFFFFFFFFFFFFFL)

enum {
      DESC_INT       = 0x00,
      DESC_DOUBLE    = 0x01,
      DESC_STRING    = 0x02,
      DESC_PAIR      = 0x03
};
#define MAXTYPES 4

// From the gc.c, Use the macro to call the error.
void noSpace(const char *msg, const char*fname, const char *func, int32_t lno);
#define outOfSpace(m) noSpace((m), __FILE__, __func__, __LINE__)
char *typeName(Generic *g);
void gc(Generic *g);
void initGC(void);

// From display.c
void display(FILE *fh, Generic *g);
void initDisplay(void);

// Int64 needs extra bytes. Macros for value and type checking.
typedef struct {
  Generic header;
  int64_t value;
} Int;
#define __int(v) (((Int*) (v))->value)
#define __isInt(v) ((v) != NULL && __descriptor(v) == DESC_INT)
Generic *newInt(int64_t v);


// Double. Macros for value, type checking and rounding to Int64.
typedef struct {
  Generic header;
  double  value;
} Double;
#define __double(v) (((Double*) (v))->value)
#define __isDouble(v) ((v) != NULL && __descriptor(v) == DESC_DOUBLE)
Generic *newDouble(double d);


// Pair. Macros for type checking,  car/cdr composites up to 3 levels, lists up to 6. This
// grabbed from JA source code, not much is used.
typedef struct {
  Generic header;
  Generic *cdr;
} Pair;
Generic *newPair(Generic *a, Generic *b);
#define cons(a, b) newPair((a), (b))
#define __isPair(v) ((v) && (__descriptor(v) == DESC_PAIR))
#define __car(v) (((Generic*) (v))->next)
#define __cdr(v) (((Pair*) (v))->cdr)
#define __caar(v) __car(__car(v))
#define __cadr(v) __car(__cdr(v))
#define __cdar(v) __cdr(__car(v))
#define __cddr(v) __cdr(__cdr(v))
#define __caaar(v) __car(__car(__car(v)))
#define __caadr(v) __car(__car(__cdr(v)))
#define __cadar(v) __car(__cdr(__car(v)))
#define __caddr(v) __car(__cdr(__cdr(v)))
#define __cdaar(v) __cdr(__car(__car(v)))
#define __cdadr(v) __cdr(__car(__cdr(v)))
#define __cddar(v) __cdr(__cdr(__car(v)))
#define __cdddr(v) __cdr(__cdr(__cdr(v)))
#define __list1(a) (cons((a), NULL))
#define __list2(a, b) (cons((a), cons((b), NULL)))
#define __list3(a, b, c) (cons((a), cons((b), cons((c), NULL))))
#define __list4(a, b, c, d) (cons((a), cons((b), cons((c), cons((d), NULL)))))
#define __list5(a, b, c, d, e) (cons((a), cons((b), cons((c), cons((d), cons((e), NULL))))))
#define __list6(a, b, c, d, e, f) (cons((a), cons((b), cons((c), cons((d), cons((e), \
             cons((f), NULL)))))))


typedef struct {
  Generic header;           // Generic header.
  char    *zterm;           // The first character of the zero terminated string.
} String;
#define __isString(v) ((v) && (__descriptor(v) == DESC_STRING))
#define __string(v) (((String *) v)->zterm)
#define __strlen(v) ((uint64_t) __next(v))
Generic *newString(uint8_t *cs, uint32_t sz);
char *string2char(Generic *s);
void initStrings(void);

Generic *assoc(char *what, Generic *l);

// See clone.c
Generic *clone(Generic *v);
void initClone(void);

// See equal.c
uint8_t equal(Generic *a, Generic *b);

// See mprotos.c
protoBlock *sub2Block(uint8_t sub);
void farmOutPayload(Generic *g);
void initProtoBufs(void);

// See position.c
Generic *lookFor(int64_t num, Generic *l);
void position(Generic *g);

// See protobuf.c
uint64_t readVARINT(void);
uint8_t *readLEN(uint32_t *sz);
void initProtobuf(void);
protoBlock *tryDifferents(uint8_t *b);
Generic *protoBuf2Generic(uint8_t *ibuf, int16_t plen, protoBlock *bl);

// See scalars.c
void initScalars(void);

// See sendmail.c
void sendMail(Generic *g);

// See user.c
void userMessage(int32_t portnum, Generic *p);
#endif


// EOF
