/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: sendmail.c
 * Author: Jed Marti
 * Description: We've parsed the message into a list structure. Format it into a message and
 *   pass off to SSMTP for mail delivery.
 * Created: Wed Dec 11 08:43:23 2024
 * Revision History:
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "mqtt.h"
#include "protobuf.h"

// Message counter. Messages saved in /tmp
int32_t smsCnt = 0;

// Noise level.
extern uint8_t Noise;

// AREDN port number for Mesh mail server. The AREDN destination node must have this port enabled
// for the mail server to operate.
extern int32_t AREDNPortnum;

// The directory to put temporary mail in. Can be changed by -s or -smsdir command line options.
extern char *SMSDir;

// Log files, see log.c, used by LOG mac ro
extern char logMsg[];
extern FILE *LOGfh;

// This is data about the user - see user.c.
extern Generic *Users;

// This is about where people are - see position.c
extern Generic *REMOTES;


      
static uint8_t isIPV4(char *mt)
/*     Check if a mail destination is an IPV4 address. If so the mail will go to a server on
 *   the remote node otherwise it goes through SSMTP.
 * Parameters:
 *   mt: A string with the destination.
 * External References: none.
 * Returned values: 1-looks like IPV4 address, 0-it doesn't.
 * Errors detected: none
 */
{
  size_t sl, i;

  // NULL string should be ignored.
  if (mt == NULL) return 0;

  // Look for only numbers and dots.
  sl = strlen(mt);
  for (i = 0; i < sl; ++i)
    if (!((mt[i] >= '0' && mt[i] <= '9') || mt[i] == '.')) return 0;
  return 1;
}



void sendMail(Generic *g)
/*     We have a generic structure from the decoded protoBuf. Get a lot of stuff we add to
 *   the message. See if this looks like a mail message. Its prolog look like:
 *      >to<from 
 *   Messages that don't look like these are ignored. Others are passed through SMTP or sent
 *   through the AREDN network to a server there.
 * Parameters:
 *   g: The parsed structure.
 * External References:
 * Returned values: none
 * Errors detected:
 */
{
  Generic *pkt, *channelid, *data, *payload, *from, *to, *fromn, *loc, *q;
  char    *pcpy;
  char    *toch, *fromch;
  char    smsfn[256];
  char    run[MAXPATHLEN];
  int32_t rc;
  FILE    *fh;
  protoBlock *pblock;
  char    fromc[128];
  double  lat, lng;

  // Show what the protoBuf and MQTT parsers found. This will be a set of nested alists.
  if (Noise > 4)
  {
    printf("**********************************************************************\n");
    display(stdout, g);
    printf("\n");
  }

  // "packet" comes from the Service Envelope.
  pkt = assoc("packet", g);
  printf("\n\npacket = "); display(stdout, pkt); printf("\n");
  printf("Users: "); display(stdout, Users);
  printf("\n\nREMOTES: "); display(stdout, REMOTES);
  printf("\n\n");
  
  // 'from' will be the last 4 bytes of the MAC address. From is an
  // integer, but we need to change it into a string to search the
  // list of users.
  from = assoc("from", pkt);
  if (__isInt(from))
  {
    sprintf(fromc, "!%0lx", __int(from));
    fromn = assoc(fromc, Users);
    if (fromn != NULL)
    {
      // Get the long name.
      fromn = assoc("long_name", fromn);
      if (fromn)
      {
        strcpy(fromc, pcpy = string2char(fromn));
        free(pcpy);
        pcpy = NULL;
      }

      // See if we have a position. Note that positions are stored by the numeric id.
      loc = lookFor(__int(from), REMOTES);
      if (loc)
      {
        q = assoc("latitude_i", loc);
        if (q) lat = ((double) __int(q)) * 1.0e-7;    // Integer to degrees.
        else lat = 999.0;                             // An illegal latitude.
        q = assoc("longitude_i", loc);
        if (q) lng = ((double) __int(q)) * 1.0e-7;    // Integer to degrees.
        else lng = 999.0;                             // An illegal longitude.
      }
    }
  }
  else sprintf(fromc, "Unknown");


  // The to field is usually -1 and the channel whatever (0=primary, 1=ARESMAIL (?), ...7-?)
  to = assoc("to", pkt);
  channelid = assoc("channel_id", g);

  // The data component contains the message descriptor and the message descriptor. We've
  // already detected that payload contains the text message. An empty messages causes an
  // early termination.
  data = assoc("data", pkt);
  payload = assoc("payload", data);
  if (payload == NULL)
  {
    if (Noise > 0) printf("%s:%s:%d Empty text message\n", __FILE__, __func__, __LINE__);
    return;
  }
  
  if (Noise > 4)
  {
    printf("  from 0x%lx\n", __int(from));
    printf("  to   0x%lx\n", __int(to));
    printf("  payload is "); display(stdout, payload);
    printf("\n");
  }

  // Remove the payload wrapping. The first character must be a > indicating who/what the
  // message is to. If not, then we display the message and punt.
  pcpy = string2char(payload);
  if (pcpy[0] != '>')
  {
    if (Noise > 4)
    {
      printf("%s:%s:%d Message payload not mail\n", __FILE__, __func__, __LINE__);
      printf("     %02X %02X %02X\n", pcpy[0], pcpy[1], pcpy[2]);
      printf("      %c   %c   %c\n", pcpy[1], pcpy[1], pcpy[2]);
    }

    // Here we could try to change some settings on the MQTT server but punt for now.

    // Display message if some noise.
    if (Noise > 0) printf("Message: %s\n", pcpy);

    return;
  }

  // Pretty weak here, all sorts of stuff could go wrong. 
  toch = expandAbbreviation(strtok(&pcpy[1], "< "));
  fromch = expandAbbreviation(strtok(NULL, " "));

  LOG("Mail to %s from %s", toch, fromch);

  // Create the file name.
  snprintf(smsfn, MAXPATHLEN, "%s/sms%04d", SMSDir, smsCnt++);
  
  if ((fh = fopen(smsfn, "wb")) == NULL)
  {
    fprintf(stderr, "%s:%s:%d Couldn't open %s for mail\n", __FILE__, __func__, __LINE__, smsfn);
    return;
  }

  // Todo:
  //   1) Look for a subject here but for now.
  //   2) Check for position of sender.
  fprintf(fh, "To: %s\n", toch);
  fprintf(fh, "Subject: Meshtastic Message\n");
  fprintf(fh, "%s\n", strtok(NULL, "\n"));
  fprintf(fh, "\nMeshtastic message from %s to %s\n", fromc, fromch);
  if (loc != NULL)  fprintf(fh, "Last known location %g, %g\n", lat, lng);
  fclose(fh);

  // Decide if AREDN or SMS.
  if (isIPV4(toch)) arednSend(toch, AREDNPortnum, smsfn);
  else
  {
    // Create command line for ssmtp.
    snprintf(run, MAXPATHLEN, "ssmtp -v %s <%s", toch, smsfn);

    // Get it going.
    if ((rc = system(run)) != 0)
    {
      fprintf(stderr, "%s:%s:%d ssmtp return code %d\n", __FILE__, __func__, __LINE__, rc);
      LOG("Message to %s rc=%d", toch, rc);
    }
  }
  
  if (Noise > 4) printf("^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^\n");

  // Free up created junk. Caller does GC on message. 
  free(pcpy);
}


// EOF
