/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: textmsg.c
 * Author: Jed Marti
 * Description: Format a text message to send to the Meshtastic mesh. Building the protoBuf
 *   packet is the main issue.
 * Created: Mon Dec 16 19:46:52 2024
 * Revision History:
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>

#include "mqtt.h"
#include "protobuf.h"

// MQTT interface handle.
extern int32_t FD;

// Log files, see log.c, used by LOG macro
extern char logMsg[];
extern FILE *LOGfh;

// A buffer for the message construction as we fake a stream.
uint8_t T[MAXMSG];
uint16_t TINDEX = 0;

// The 4 byte MAC address used to identify the client the message is from.
extern uint32_t clientID;

// We increment this on each message.
uint32_t messageNumber = 0;



static void tputch(uint8_t v)
/*     Put a single byte into the buffer.
 * Parameters:
 *   v: What to put there.
 * External References:
 *   TINDEX: index into the buffer.
 *   T[]: A stream buffer.
 * Returned values: none
 * Errors detected: buffer overflow.
 */
{
  if (TINDEX >= MAXMSG)
  {
    fprintf(stderr, "%s:%s:%d protoBuf size %d exceeded\n", __FILE__, __func__, __LINE__, MAXMSG);
    LOG("protoBuf size %d exceeded", MAXMSG);
    exit(-1);
  }
  T[TINDEX++] = v;
}



static void tput32(uint32_t v)
/*     Little endian integer to buffer.
 * Parameters:
 *   v: unsigned 32 bitter.
 * External References: none
 * Returned values: none
 * Errors detected: lower down.
 */
{
  tputch(v & 0xFF);
  tputch((v >> 8) & 0xFF);
  tputch((v >> 16) & 0xFF);
  tputch((v >> 24) & 0xFF);
}


static void tputString(char *str)
/*     For short strings only (< 128). Strings needing more done by hand.
 * Parameters:
 *   str: 0 terminated string to write.
 * External References: none
 * Returned values: none
 * Errors detected: lower down. Should probably check string length as length is VARINT.
 */
{
  int32_t sl, i;
  sl = strlen(str);
  tputch(sl);
  for (i = 0; i < sl; ++i) tputch(str[i]);
}



void text2protoBuf(char *msg)
{
  int32_t ssl, sl;
  int32_t dl, ml;
  int32_t i;
  char    gateway[10];

  // msg is a 0 terminated string. Length is with the flags and length.
  ssl = strlen(msg);
  sl = ssl + 2;
  if (ssl > 127) ++sl;

  // The data length.
  dl = 2 +                // Portnum.
       ssl +               // payload + it's length.
       2;                 // Terminating 0.
  
  
  // The mesh length.
  ml = 5 +                // from = 1, I32
       5 +                // to = 2, I32
       2 +                // channel = 3, VARINT *
       2 + dl +           // data = 4
       5 +                // id = 6,  I32 *
       5 +                // rx_time = 7, I32*
       6;                 // hop_limit, hop_start = 9, 14, 15

  TINDEX = 0;

  printf("%s:%s:%d ssl=%d, sl=%d, dl=%d, ml=%d\n", __FILE__, __func__, __LINE__, ssl, sl,  dl, ml);
  
  // Service packet. Indicates MeshPacket follows and length given is a VARINT.
  tputch((1 << 3) | LEN);
  if (ml > 127)
  {
    tputch((ml & 0x7F) | 0x80);
    tputch(ml >> 7);
  }
  else tputch(ml);

    // Mesh packet now.
    // fixed32 from = 1.
    tputch((1 << 3) | I32);
    tput32(-1);

    // fixed32 to = 2 - shouled be last 4 bytes of clientid.
    tputch((2 << 3) | I32);
    tput32(clientID);

    // varint channel = 3
    tputch((3 << 3) | VARINT);
    //    tputch(2);                     // Channel 2 is ARESMAIL.
    tputch(1);                         // Try LongSlow for Now.
    
      // Data group
    tputch((4 << 3) | LEN);
      if (dl > 127)
      {
        tputch((dl & 0x7F) | 0x80);
        tputch(dl >> 7);
      }
      else tputch(dl);

      // Portnum enum = 1, we're sending a text mnessage.
      tputch((1 << 3) | VARINT);
      tputch(TEXT_MESSAGE_APP);

      // This is the text message.
      tputch((2 << 3) | LEN);
      if (ssl > 127)
      {
        tputch((ssl & 0x7F) | 0x80);
        tputch(ssl >> 7);
      }
      else tputch(ssl);
      for (i = 0; i < ssl; ++i) tputch(msg[i]);

      // Bitfield is 0 a funky way to terminate a string. Indicates we're OK?
      tputch(0x48);
      tputch(0);
  
    // Remainder of Mesh packet.
      // id=6, I32. Give our number which is last 4 digits + our index.
      tputch(6 << 3 | I32);
      tput32(((messageNumber ++) << 16) + clientID);

      // Time since 1970.
      tputch(7 << 3 | I32);
      tput32(time(0));

    // The hop limit, want_ack, via_mqtt, hop_start
    tputch((9 << 3) | VARINT);
    tputch(3);
    //    tputch((10 << 3) | VARINT);  want_ack (it works).
    //    tputch(1);
    tputch((14 << 3) | VARINT);
    tputch(1);
    tputch((15 << 3) | VARINT);
    tputch(0);
    printf("%s:%s:%d After data, TINDEX=%d\n", __FILE__, __func__, __LINE__, TINDEX);

  // Remainder of ServiceEnvelope. Channel_ID.
  tputch((2 << 3) | LEN);
  //  tputString("ARESMAIL");
  tputString("LongSlow");

  // Sending from 
  tputch((3 << 3) | LEN);
  sprintf(gateway, "!%08x", clientID);
  tputString(gateway);
}



void publishMessage(char *msg)
{
  printf("%s:%s:%d Converting to protoBuf '%s'\n", __FILE__, __func__, __LINE__, msg);
  text2protoBuf(msg);
  printf("%s:%s:%d Sending %d\n", __FILE__, __func__, __LINE__, TINDEX);
  mqtt2Mesh(T, TINDEX);
}
  
  

// EOF
