/* Copyright 2024 by COG9 Enterprises LLC
 * All rights reserved.
 * Title: user.c
 * Author: Jed Marti
 * Description: Parse the information in a user message. We're most interested in their device,
 *   long name, and short name.
 * Created:
 * Revision History:
 *   Mon Dec 30 07:35:17 2024 - Code review.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "mqtt.h"
#include "protobuf.h"

// For debugging.
extern uint8_t Noise;

// The control block for the User data record.
extern protoBlock protoUser;

// A list of user records
Generic *Users = NULL;

// Log files, see log.c, used by LOG macro
extern char logMsg[];
extern FILE *LOGfh;

// The hardware supported or at least listed.
static char *HardwareModel[] = {
                                "UNSET",                           // = 0
                                "TLORA_V2",                        // = 1
                                "TLORA_V1",                        // = 2
                                "TLORA_V2_1_1P6",                  // = 3
                                "TBEAM",                           // = 4
                                "HELTEC_V2_0",                     // = 5
                                "TBEAM_V0P7",                      // = 6
                                "T_ECHO",                          // = 7
                                "TLORA_V1_1P3",                    // = 8
                                "RAK4631",                         // = 9
                                "HELTEC_V2_1",                     // = 10
                                "HELTEC_V1",                       // = 11
                                "LILYGO_TBEAM_S3_CORE",            // = 12
                                "RAK11200",                        // = 13
                                "NANO_G1",                         // = 14
                                "TLORA_V2_1_1P8",                  // = 15
                                "TLORA_T3_S3",                     // = 16
                                "NANO_G1_EXPLORER",                // = 17
                                "NANO_G2_ULTRA",                   // = 18
                                "LORA_TYPE",                       // = 19
                                "WIPHONE",                         // = 20
                                "WIO_WM1110",                      // = 21
                                "RAK2560",                         // = 22
                                "HELTEC_HRU_3601",                 // = 23
                                "HELTEC_WIRELESS_BRIDGE",          // = 24
                                "STATION_G1",                      // = 25
                                "RAK11310",                        // = 26
                                "SENSELORA_RP2040",                // = 27
                                "SENSELORA_S3",                    // = 28
                                "CANARYONE",                       // = 29
                                "RP2040_LORA",                     // = 30
                                "STATION_G2",                      // = 31
                                "LORA_RELAY_V1",                   // = 32
                                "NRF52840DK",                      // = 33
                                "PPR",                             // = 34
                                "GENIEBLOCKS",                     // = 35
                                "NRF52_UNKNOWN",                   // = 36
                                "PORTDUINO",                       // = 37
                                "ANDROID_SIM",                     // = 38
                                "DIY_V1",                          // = 39
                                "NRF52840_PCA10059",               // = 40
                                "DR_DEV",                          // = 41
                                "M5STACK",                         // = 42
                                "VHELTEC_V3",                      // = 43;
                                "HELTEC_WSL_V3",                   // = 44;
                                "BETAFPV_2400_TX",                 // = 45;
                                "BETAFPV_900_NANO_TX",             // = 46;
                                "RPI_PICO",                        // = 47;
                                "HELTEC_WIRELESS_TRACKER",         // = 48;
                                "HELTEC_WIRELESS_PAPER",           // = 49;
                                "T_DECK",                          // = 50;
                                "T_WATCH_S3",                      // = 51;
                                "PICOMPUTER_S3",                   // = 52;
                                "HELTEC_HT62",                     // = 53;
                                "EBYTE_ESP32_S3",                  // = 54;
                                "ESP32_S3_PICO",                   // = 55;
                                "CHATTER_2",                       // = 56;
                                "HELTEC_WIRELESS_PAPER_V1_0",      // = 57;
                                "HELTEC_WIRELESS_TRACKER_V1_0",    // = 58;
                                "UNPHONE",                         // 59;
                                "TD_LORAC",                        // 60;
                                "CDEBYTE_EORA_S3",                 // 61;
                                "TWC_MESH_V4",                     // 62;
                                "NRF52_PROMICRO_DIY",              // 63;
                                "RADIOMASTER_900_BANDIT_NANO",     // 64;
                                "HELTEC_CAPSULE_SENSOR_V3",        // 65;
                                "HELTEC_VISION_MASTER_T190",       // 66;
                                "HELTEC_VISION_MASTER_E213",       // 67;
                                "HELTEC_VISION_MASTER_E290",       // 68;
                                "HELTEC_MESH_NODE_T114",           // 69;
                                "SENSECAP_INDICATOR",              // 70;
                                "TRACKER_T1000_E",                 // 71;
                                "RAK3172",                         // 72;
                                "WIO_E5",                          // 73;
                                "RADIOMASTER_900_BANDIT",          // 74;
                                "ME25LS01_4Y10TD",                 // 75;
                                "RP2040_FEATHER_RFM95",            // 76;
                                "M5STACK_COREBASIC",               // 77;
                                "M5STACK_CORE2",                   // 78;
                                "RPI_PICO2",                       // 79;
                                "M5STACK_CORES3",                  // 80;
                                "SEEED_XIAO_S3",                   // 81;
                                "MS24SF1",                         // 82;
                                "TLORA_C6"};                       // 83;



static void replaceUser(Generic *uid, Generic *new)
/*     Look for a previous instance of this user. If found, replace the information. 
 *   The caller will have checked. 
 * Parameters:
 *   uid: The user's id (last bytes of MAC address).
 *   new: The new info.
 * External References:
 *   Users: An association list of the users.
 * Returned Values: none
 * Errors detected: none
 */
{
  Generic *p;
  for (p = Users; __isPair(p); p = __cdr(p))
  {
    if (equal(__caar(p), uid))
    {
      // p will be (id  (a . b) (c . d) ...)
      p = __car(p);

      gc(__cdr(p));
      __cdr(p) = new;
      return;
    }
  }
}



void userMessage(int32_t portnum, Generic *p)
/*     g is from PUBLISH and the Service envelope with the NODEINFO_APP portnum but the User
 *   protoBuf description. Get all the user information and store on the user list. The caller
 *   will free the message.
 * Parameter:
 *   portnum: The protoBuf record number.
 *   p: The parsed version of this record (the User record, 
 * External References:
 *   HardwareModel[]: Names of various bits of hardware maybe supporting Meshtastic.
 *   Users: A list of users we've encountered. Each identified by their id and associated with
 *      (id . <protoBuf User message)
 * Returned values: none, function modifies Users association list.
 * Errors Detected: none here, not very safe.
 */
{
  Generic *payload, *user, *id, *longname, *shortname, *hwmodel;
  Generic *old;
  char    *slongname, *sshortname;
  int32_t ihwmodel;
  char    shwmodel[64];

  if (Noise > 5) printf("%s:%s:%d User message", __FILE__, __func__, __LINE__);

  // Some useful stuff in the top level package (maybe) so the complete structure gets sent
  // every time.
  payload = assoc("packet", p);
  payload = assoc("data", payload);
  payload = assoc("payload", payload);

  // We have the User string, convert it to a structure.
  user = protoBuf2Generic(__string(payload), __strlen(payload), &protoUser);
  if (Noise > 1)
  {
    printf("%s:%s:%d User contents ", __FILE__, __func__, __LINE__);
    display(stdout, user);
    printf("\n");
  }
  
  // We use the id as the user attribute (last bits of MAC address).
  id = assoc("id", user);

  longname = assoc("long_name", user);
  if (__isString(longname)) slongname = string2char(longname);
  else slongname = NULL;
  
  shortname = assoc("short_name", user);
  if (__isString(shortname)) sshortname = string2char(shortname);
  else sshortname = NULL;
  
  hwmodel = assoc("hw_model", user);
  if (__isInt(hwmodel))
  {
    ihwmodel = __int(hwmodel);
    if (ihwmodel < 0 || ihwmodel > sizeof(HardwareModel) / sizeof(char*))
      snprintf(shwmodel, 64, "model=%d", ihwmodel);
    else
      strncpy(shwmodel, HardwareModel[ihwmodel], 64);
  }
  else snprintf(shwmodel, 64, "???");
  
  // LOG stuff.
  LOG("Name %s, %s, model %s", slongname, sshortname, shwmodel);
  
  // If not already there, then add to front of list cloning it so we can release message
  // structure.
  printf("Looking for '%s' on Users\n", string2char(id));
  if ((old = assoc(__string(id), Users)) == NULL)
    Users = cons(cons(id, clone(user)), Users);

  // Otherwise, replace the current one.
  else replaceUser(id, clone(user));

  // Clean up storage.
  if (sshortname) free(sshortname);
  if (slongname) free(slongname);

}


// EOF
